local lovelive = {
  version = 'v0.1.0',
}

local options = {
  panelWidth = 350,
  panelColor = { 0, 0, 0, 180 },
  panelSlideSpeed = 10,
  panelFadeSpeed = 5,
  panelScrollSpeed = 10,
  panelScrollDistance = 80,

  scrollbarWidth = 3,
  scrollbarSpacingLeft = 10,
  scrollbarSpacingTop = 10,
  scrollbarColor = { 255, 255, 255, 180 },

  groupFont = nil,
  groupFontSize = 32,

  valueFont = nil,
  valueFontSize = 24,

  textColor = { 255, 255, 255 },
  textHoverColor = { 100, 100, 255 },

  paddingLeft = 30,
  paddingTop = 30,

  groupSpacing = 25,
  valueSpacing = 10,
}

local panel = {}
local elements = {}
local resources = {}
local input = {}
local lastAddedValues = {}

local function noop() end

local function clamp(n, min, max)
  return n < min and min or n > max and max or n
end

local function lerp(value, target, delta)
  return value + (target - value) * math.min(delta, 1)
end

local function colorOpacity(color, opacity)
  return color[1], color[2], color[3], opacity * (color[4] or 255)
end


local function areaFromBounds(left, top, right, bottom)
  return {
    left = left,
    top = top,
    right = right,
    bottom = bottom,
  }
end

local function areaFromRect(x, y, width, height)
  return {
    left = x,
    top = y,
    right = x + width,
    bottom = y + height,
  }
end

local function getAreaRect(area)
  return area.left, area.top, area.right - area.left, area.bottom - area.top
end

local function areaContains(area, x, y)
  return x > area.left and x < area.right
  and y > area.top and y < area.bottom
end


local function addElement(text, font, spacing, info)
  local x = options.paddingLeft
  local y = options.paddingTop

  if #elements > 0 then
    y = elements[#elements].area.bottom + spacing
  end

  local element = {
    text = text,
    font = font,
    area = areaFromRect(x, y, font:getWidth(text), font:getHeight(text)),
    info = info or {},
  }

  table.insert(elements, element)
  return element
end

local function addValueElement(...)
  local element = addElement(...)
  element.info.type = 'value'
  element.info.whileChanged = {}
  element.info.afterChanged = {}
  table.insert(lastAddedValues, element)
  return element
end


function lovelive.init(userOptions)
  lovelive.configure(userOptions)

  panel.area = areaFromRect(-options.panelWidth, 0, options.panelWidth, love.graphics.getHeight())
  panel.opacity = 0
  panel.visible = false
  panel.scroll = 0
  panel.scrollTarget = 0
  panel.maxScroll = 0

  panel.scrollbarOpacity = 0
  panel.scrollbarTimer = 0

  resources.groupFont = options.groupFont
    and love.graphics.newFont(options.groupFont, options.groupFontSize)
    or love.graphics.newFont(options.groupFontSize)

  resources.valueFont = options.valueFont
    and love.graphics.newFont(options.valueFont, options.valueFontSize)
    or love.graphics.newFont(options.valueFontSize)

  local mx, my = unpack(love.mouse and { love.mouse.getPosition() } or { 0, 0 })
  input.x = mx
  input.y = my
  input.state = 'idle'
  input.activeElement = nil
  input.usingMouse = false

  return lovelive
end

function lovelive.configure(userOptions)
  for k,v in pairs(userOptions or {}) do
    options[k] = v
  end
  return lovelive
end

function lovelive.visible(isVisible)
  panel.visible = isVisible
  return lovelive
end

function lovelive.toggle()
  panel.visible = not panel.visible
  return lovelive
end

function lovelive.group(title)
  addElement(title, resources.groupFont, options.groupSpacing)
  return lovelive
end

function lovelive.number(name, object, field, increment, min, max)
  lastAddedValues = {}

  addValueElement(name .. ': ' .. object[field], resources.valueFont, options.valueSpacing, {
    valueType = 'number',
    object = object,
    field = field,
    increment = increment or 1,
    min = min or -math.huge,
    max = max or math.huge,
  })
  return lovelive
end

function lovelive.boolean(name, object, field)
  lastAddedValues = {}

  addValueElement(name .. ': ' .. tostring(object[field]), resources.valueFont, options.valueSpacing, {
    valueType = 'boolean',
    object = object,
    field = field,
  })
  return lovelive
end

function lovelive.action(name, func)
  lastAddedValues = {}

  addValueElement(name, resources.valueFont, options.valueSpacing, {
    valueType = 'action',
    action = func,
  })
  return lovelive
end

function lovelive.object(objectName, object, inc, min, max)
  lastAddedValues = {}

  local function newObject(objectName, object, cache)
    for field, value in pairs(object) do
      local name
      if type(field) == 'string' and field:find('^[%a_][%w_]-$') then
        name = objectName .. '.' .. field
      else
        name = objectName .. '[' .. tostring(field) .. ']'
      end

      if type(value) == 'number' then
        lovelive.number(name, object, field, inc, min, max)
      elseif type(value) == 'boolean' then
        lovelive.boolean(name, object, field)
      elseif type(value) == 'function' then
        lovelive.action(name, value)
      elseif type(value) == 'table' then
        if not cache[value] then
          cache[value] = true
          newObject(name, value, cache)
        else
          cache[value] = nil
        end
      end
    end
  end

  newObject(objectName, object, {})

  return lovelive
end

function lovelive.whileChanged(callback)
  for i, element in ipairs(lastAddedValues) do
    table.insert(element.info.whileChanged, callback)
  end
  return lovelive
end

function lovelive.afterChanged(callback)
  for i, element in ipairs(lastAddedValues) do
    table.insert(element.info.afterChanged, callback)
  end
  return lovelive
end

function lovelive.updateUI(dt)
  local x = panel.visible and 0 or -options.panelWidth
  local panelHeight = love.graphics.getHeight()
  local opacity = panel.visible and 1 or 0

  -- update panel position
  panel.area = areaFromRect(
    lerp(panel.area.left, x, dt * options.panelSlideSpeed), 0,
    options.panelWidth, panelHeight)

  -- update panel opacity
  panel.opacity = lerp(panel.opacity, opacity, dt * options.panelFadeSpeed)

  -- update scrolling stuff
  panel.maxScroll = math.max(elements[#elements].area.bottom + options.groupSpacing - panelHeight, 0)
  if panel.maxScroll ~= 0 then
    panel.scroll = lerp(panel.scroll, panel.scrollTarget, dt * options.panelScrollSpeed)

    if panel.scroll < 0 then
      panel.scroll = 0
      panel.scrollTarget = 0
    elseif panel.scroll > panel.maxScroll then
      panel.scroll = panel.maxScroll
      panel.scrollTarget = panel.maxScroll
    end
  else
    panel.scroll = 0
  end

  panel.scrollbarTimer = panel.scrollbarTimer - dt
  if panel.scrollbarTimer > 0 then
    panel.scrollbarOpacity = lerp(panel.scrollbarOpacity, 1, dt * 8)
  else
    panel.scrollbarOpacity = lerp(panel.scrollbarOpacity, 0, dt * 5)
  end

  return lovelive
end

function lovelive.updateInput(isDown, mx, my)
  my = my + panel.scroll

  local dx, dy = mx - input.x, my - input.y

  if not panel.visible then
    if input.state ~= 'idle' then
      input.state = 'idle'
    end
    return
  end

  if input.state ~= 'busy' then
    input.activeElement = nil
    for i, element in ipairs(elements) do
      if areaContains(element.area, mx, my)
      and element.info.type == 'value' then
        input.activeElement = element
        break
      end
    end
  end

  local active = input.activeElement
  if input.state == 'idle' then
    if active and not isDown then
      input.state = 'hovering'
    end
  elseif input.state == 'hovering' then
    if not active then
      input.state = 'idle'
    elseif isDown then
      if active.info.valueType == 'boolean' then
        active.info.object[active.info.field] = not active.info.object[active.info.field]
      elseif active.info.valueType == 'action' then
        active.info.action()
      end
      input.state = 'busy'
    end
  elseif input.state == 'busy' then
    if isDown then
      if active and active.info.valueType == 'number'
      and (dx ~= 0 or dy ~= 0) then
        local info = active.info
        local object, field = info.object, info.field
        local delta = math.abs(dx) > math.abs(dy) and dx or -dy
        object[field] = clamp(object[field] + info.increment * delta, info.min, info.max)

        for i, callback in ipairs(active.info.whileChanged) do
          callback(object[field], field, object)
        end
      end
    else
      input.state = active and 'hovering' or 'idle'

      local info = active.info
      for i, callback in ipairs(info.afterChanged) do
        callback(info.object[info.field], info.field, info.object)
      end
    end
  end

  if input.usingMouse then
    local cursor = nil
    local relative = false
    if input.state == 'hovering' then
      cursor = resources.hoverCursor
    elseif input.state == 'busy'
    and active and active.info.valueType == 'number' then
      relative = true
    end

    love.mouse.setCursor(cursor)
    love.mouse.setRelativeMode(relative)
  end

  input.x, input.y = mx, my

  return lovelive
end

function lovelive.update(dt, isDown, mx, my)
  lovelive.updateUI(dt)
  if input.usingMouse then
    lovelive.updateInput(love.mouse.isDown(1), love.mouse.getPosition())
  else
    lovelive.updateInput(isDown, mx, my)
  end
end

function lovelive.getInputState()
  return input.state
end

function lovelive.scroll(scroll, mx, my)
  if areaContains(panel.area, mx, my)
  or mx == nil and my == nil then
    panel.scrollTarget = panel.scrollTarget - scroll * options.panelScrollDistance
    panel.scrollbarTimer = 0.5
  end
end

function lovelive.draw()
  if panel.opacity > 0.05 then
    love.graphics.push('all')
    love.graphics.origin()
    love.graphics.setScissor(getAreaRect(panel.area))

    love.graphics.setColor(colorOpacity(options.panelColor, panel.opacity))
    love.graphics.rectangle('fill', getAreaRect(panel.area))

    local pageHeight = (panel.area.bottom - panel.area.top) - options.scrollbarSpacingTop * 2
    local scrollbarHeight = pageHeight / (panel.maxScroll + pageHeight) * pageHeight
    local scrollDistance = pageHeight - scrollbarHeight
    local scrollRatio = panel.scroll / panel.maxScroll

    love.graphics.setColor(colorOpacity(options.scrollbarColor, panel.scrollbarOpacity * panel.opacity))
    love.graphics.rectangle('fill', options.scrollbarSpacingLeft,
      options.scrollbarSpacingTop + scrollRatio * scrollDistance, options.scrollbarWidth,
      scrollbarHeight)

    love.graphics.translate(panel.area.left, panel.area.top - panel.scroll)

    for i, element in ipairs(elements) do
      love.graphics.setFont(element.font)
      love.graphics.setColor(colorOpacity(options.textColor, panel.opacity))
      love.graphics.print(element.text, element.area.left, element.area.top)
    end

    love.graphics.pop()
  end

  elements = {}

  return lovelive
end


function lovelive.useMouse()
  local wheelmoved = love.wheelmoved or noop
  function love.wheelmoved(dx, dy)
    lovelive.scroll(dy, love.mouse.getPosition())
    wheelmoved(dx, dy)
  end

  resources.hoverCursor = love.mouse and love.mouse.getSystemCursor('hand')
  input.usingMouse = true

  return lovelive
end

function lovelive.useToggleKey(toggleKey)
  local keypressed = love.keypressed or noop
  function love.keypressed(key, ...)
    if key == toggleKey then
      lovelive.toggle()
    end
    keypressed(key, ...)
  end

  return lovelive
end

return lovelive
